# -*- coding: utf-8 -*-
import re

CAP_RE = re.compile('(.)([A-Z])')


def to_pascal_case(name):
    """
    Turn a snake_case name into PascalCase.
    """
    return name.title().replace("_", "")


def from_pascal_case(name):
    """
    Turn a PascalCase name into snake_case.
    """
    return CAP_RE.sub(r'\1_\2', name).lower()


def to_camel_case(name):
    """
    Turn a snake_case name into camelCase.
    """
    output = name.title().replace("_", "")
    return output[0].lower() + output[1:]


def from_camel_case(name):
    """
    Turn a camelCase name into snake_case.
    """
    return from_pascal_case(name)



__title__ = 'mew'
__description__ = 'Python dataclass serializer/deserializer'
__url__ = 'http://github.com/cliffxuan/mew'
__version__ = '0.0.5'
__author__ = 'Cliff Xuan'
__author_email__ = 'cliff.xuan@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2019 Cliff Xuan'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
"""
from dataclasses import is_dataclass
import datetime as dt
import enum
import typing
import uuid

from mew.serializer import dumps, loads


class NotSupported(Exception):
    """not supported"""

    def __init__(self, types, message):
        self.types = types
        self.message = message


scalar_types = [str, int, float, bool, type(None)]
directly_supported_types = [
    uuid.UUID,
    dt.datetime,
    *scalar_types,
]


def is_namedtuple(t):
    b = t.__bases__
    if len(b) != 1 or b[0] != tuple:
        return False
    f = getattr(t, "_fields", None)
    if not isinstance(f, tuple):
        return False
    return all(type(n) == str for n in f)


def find_unsupported(t) -> typing.List[typing.Any]:
    if t in directly_supported_types:
        return []
    if isinstance(t, type):  # if it's a class
        if issubclass(t, enum.Enum):
            return []
        if is_dataclass(t):
            unsupported = [
                v.type
                for v in t.__dataclass_fields__.values()
                if find_unsupported(v.type)
            ]
            return unsupported
        if is_namedtuple(t):
            unsupported = [
                v
                for v in t._field_types.values()
                if find_unsupported(v)
            ]
            return unsupported
    if hasattr(t, "__origin__"):  # if it's a type in typing module
        origin = t.__origin__
        if origin == typing.Union:
            unsupported = [
                arg
                for arg in t.__args__
                if find_unsupported(arg)
            ]
            return unsupported
        if origin == list:
            return find_unsupported(t.__args__[0])
        # TODO covert more of typing.XXX
        return [t]
    return [t]


def serializable(t):
    """adds dumps() and loads() to the class"""
    unsupported_types = find_unsupported(t)
    if unsupported_types:
        # TODO specified which types are not supported
        raise NotSupported(
            unsupported_types, f"unsupported type {unsupported_types}"
        )
    t.dumps = dumps
    t.loads = classmethod(loads)
    return t



# -*- coding: utf-8 -*-
import datetime as dt
import enum
import json
import uuid
from collections.abc import Collection, Mapping
from typing import Any, Callable, Union

from dataclasses import asdict, is_dataclass

UTC = dt.timezone.utc


def serialize(
    obj: Any,
    convert_key: Callable = lambda x: x,
) -> Any:
    """
    Takes an object and returns a representation of that object
    that can be json encoded. If no special behaviour is known
    for the object then the object itself will be returned unless
    a default conversion function has been provided.
    """
    # Format as "10fb6968-6b54-44c8-9365-e0b3934ae156".
    if isinstance(obj, uuid.UUID):
        return str(obj)

    # Format datetimes as ISO-8601 w/ timezone.
    if isinstance(obj, dt.datetime):
        return obj.isoformat()

    # Serialise enums to their value.
    if isinstance(obj, enum.Enum):
        return obj.value

    # Recursively serialise dictionaries.
    if isinstance(obj, Mapping):
        # TODO this is assuming k is a string
        return {
            convert_key(k): serialize(v, convert_key)
            for k, v in obj.items()
        }

    # Convert dataclass instance as dict
    if is_dataclass(obj) and not isinstance(obj, type):
        return serialize(asdict(obj), convert_key)

    # Recursively serialise stuff that looks like a list or set.
    if isinstance(obj, Collection) and not isinstance(obj, str):
        return [serialize(i, convert_key) for i in obj]

    return obj


# TODO type annotate t
# TODO code example to use
def deserialize(t, value, convert_key: Callable = lambda x: x):
    if t in [str, int, float, bool, type(None)]:
        return value

    if t == uuid.UUID:
        return uuid.UUID(value)

    if t == dt.datetime:
        return dt.datetime.fromisoformat(value)

    if is_dataclass(t):
        return t(
            **{
                # TODO parameterize from_pascal_case
                convert_key(k): deserialize(
                    t.__dataclass_fields__[convert_key(k)].type,
                    v,
                    convert_key
                )
                for k, v in value.items()
            }
        )

    if isinstance(t, type):
        if issubclass(t, enum.Enum):
            for item in t:
                if item.value == value:
                    return item
            raise Exception(
                f"cannot find an item in Enum {t} with value {value}"
            )
        if issubclass(t, tuple):
            return t(*value)

    if hasattr(t, "__origin__") and hasattr(t, "__args__"):
        origin = t.__origin__
        args = t.__args__
        if origin == list:
            return [deserialize(args[0], item, convert_key) for item in value]
        if origin == Union:
            for arg in args:
                try:
                    return deserialize(arg, value, convert_key)
                except TypeError:
                    pass
            raise Exception(
                f"cannot convert type {t} for value {value}"
            )
        # TODO convert more of type.XXX
    raise Exception(f"unsupported type {t} for value {value}")


def dumps(o, *args, **kwargs):
    convert_key = kwargs.pop('convert_key', lambda x: x)
    serialized = serialize(o, convert_key=convert_key)
    return json.dumps(serialized, *args, **kwargs)


def loads(t, data, convert_key: Callable = lambda x: x):
    data_asdict = json.loads(data)
    return deserialize(t, data_asdict, convert_key)



#!/usr/bin/env python
import os
import sys

from codecs import open

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

# 'setup.py publish' shortcut.
if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist bdist_wheel')
    os.system('twine upload dist/*')
    sys.exit()

packages = ['mew']

requires = []
test_requirements = [
    'hypothesis',
    'dateutils',
    'flake8',
    'pytest-cov',
    'pytest>=2.8.0'
]

about = {}
with open(os.path.join(here, 'mew', '__version__.py'), 'r', 'utf-8') as f:
    exec(f.read(), about)

with open('README.md', 'r', 'utf-8') as f:
    readme = f.read()

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    packages=packages,
    package_data={'': ['LICENSE', 'NOTICE']},
    package_dir={'mew': 'mew'},
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=requires,
    license=about['__license__'],
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython'
    ],
    tests_require=test_requirements
)
